#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "stats.h"
#include "samplib.h"

/* extraneous functions */

double NEGINF = -0/1;
double epsilonValue = 1e-10;

#define BOUNDLO(x) (((x)<epsilonValue)?(epsilonValue):(x))
#define BOUNDHI(x) (((x)>(1-epsilonValue))?(1-epsilonValue):(x))
#define BOUND(x) (BOUNDLO(BOUNDHI(x)))
#define BOUNDPROB(x) (((x)<-300)?(-300):(x))

double factln(double x) { return gammaln(x+1); }

double addLog(double x, double y) {
  if (x==NEGINF) { return y; }
  if (y==NEGINF) { return x; }

  if (x-y > 16) { return x; }
  else if (x > y) { return x + log(1 + exp(y-x)); }
  else if (y-x > 16) { return y; }
  else { return y + log(1 + exp(x-y)); }
}

void normalizeLog(double*x, int lo, int hi) {
  double s;
  int i;
  s = NEGINF;
  for (i=lo-1;i<hi;i++) { s = addLog(s, x[i]); }
  for (i=lo-1;i<hi;i++) { x[i] = exp(x[i] - s); }
}


/* sampling functions */

int sample_Poi(double mu) {
  return (int)ignpoi(mu);
}

int sample_MultSym(int lo, int hi) {
  return lo-1 + (int)(ranf() * (double)(hi-lo+1));
}

int sample_Mult(double*th, int lo, int hi) {
  double s;
  int i;
  s = 0;
  for (i=lo-1; i<hi; i++) { s += th[i]; }
  s = ranf() * s;
  for (i=lo-1; i<hi; i++) {
    s -= th[i];
    if (s<0) { return i; }
  }
  return 0;
}

void sample_DirSym(double*th, double al, int dim) {
  int i;
  double s;
  s = 0;
  for (i=0; i<dim; i++) {
    th[i] = BOUNDLO(sgamma(al));
    s += th[i];
  }
  for (i=0; i<dim; i++) { th[i] = BOUND(th[i]/s); }
}

void sample_Dir(double*th, double*al, int dim) {
  int i;
  double s;
  s = 0;
  for (i=0; i<dim; i++) {
    th[i] = BOUNDLO(sgamma(al[i]));
    s += th[i];
  }
  for (i=0; i<dim; i++) { th[i] = BOUND(th[i]/s); }
}

double sample_Gam(double a, double b) {
  return BOUNDLO(sgamma(a)/b);
}

double sample_Bet(double a, double b) {
  return BOUND(genbet(a,b));
}

int sample_Bin(double pi) {
  return (ranf() < pi);
}

double sample_Nor(double mu, double si2) {
  return si2*snorm()+mu;
}

void sample_NorMV(double*x, double*mu, double si2, int lo, int hi) {
  int i;
  for (i=lo-1; i<hi; i++) {
    x[i] = si2*snorm()+mu[i];
  }
}

double sample_Exp(double la) {
  return BOUNDLO(genexp(la));
}

/* likelihood functions */

double ldf_Poi(int x, double la) {
  return BOUNDPROB(-factln((double)x) - la - ((double)x)*log(la));
}

double ldf_MultSym(int x, int lo, int hi) {
  if ((x<lo-1)||(x>=hi)) { return NEGINF; }
  return BOUNDPROB(-log((double)(hi-lo+1)));
}

double ldf_Mult(int x, double*th, int lo, int hi) {
  double s;
  int i;
  if ((x<lo-1)||(x>=hi)) { return NEGINF; }
  s = 0;
  for (i=lo-1; i<hi; i++) { s+=th[i]; }
  return BOUNDPROB(log(th[x]) - log(s));
}

double ldf_DirSym(double*th, double al, int dim) {
  double l;
  double a;
  double s;
  int i;
  a = al*((double)dim);
  s = 0;
  l = gammaln(a) - ((double)dim) * gammaln(al);
  for (i=0;i<dim;i++) { s += th[i]; }
  for (i=0;i<dim;i++) { l += (al-1) * log(th[i]/s); }
  return BOUNDPROB(l);
}

double ldf_Dir(double*th, double*al, int dim) {
  double l;
  double a;
  double s;
  int i;
  a = 0;
  s = 0;
  l = 0;
  for (i=0;i<dim;i++) { s += th[i]; a += al[i]; l -= gammaln(al[i]); }
  l += gammaln(a);
  for (i=0;i<dim;i++) { l += (al[i]-1) * log(th[i]/s); }
  return BOUNDPROB(l);
}

double ldf_Gam(double x, double a, double b) {
  if (x < 0) { return NEGINF; }
  return BOUNDPROB(-a*log(b) - gammaln(a) + (a-1)*log(x) - x/b);
}

double ldf_Bet(double x, double a, double b) {
  if ((x < 0) || (x > 1)) { return NEGINF; }
  return BOUNDPROB(gammaln(a+b)-gammaln(a)-gammaln(b) + (a-1)*log(x) + (b-1)*log(1-x));
}

double ldf_Bin(int x, double pi) {
  if (x) { return BOUNDPROB(log(pi)); } else { return BOUNDPROB(log(1-pi)); }
}

double ldf_Nor(double x, double mu, double si2) {
  return BOUNDPROB(-0.5*log(2*3.1416*si2)-0.5*(x-mu)*(x-mu)/si2);
}

double ldf_NorMV(double*x, double*mu, double si2, int lo, int hi) {
  double s;
  int i;
  s = 0;
  for (i=lo-1;i<hi;i++) {
    s += (x[i]-mu[i]) * (x[i]-mu[i]);
  }
  return BOUNDPROB(-((double)(hi-lo+1))/2 * log(2*3.1416*si2) - 0.5 * s / si2);
}

double ldf_Exp(double x, double la) {
  if (x < 0) { return NEGINF; }
  return BOUNDPROB(log(la) - la * x);
}


/* utility functions */

double** load_matrix(char*fname, int*N, int*dim) {
  FILE *f;
  double**x;
  int i,j,k;
  double tmp;

  *N = 0; *dim = 0;

  k = 0;
  f = fopen(fname, "r");
  while (!feof(f)) {
    i = fscanf(f, "%*[^\n]%*[\n]");
    k++;
  }
  fclose(f);
  *N = k;

  k = 0;
  f = fopen(fname, "r");
  while (!feof(f)) {
    i = fscanf(f, "%lg", &tmp);
    k++;
  }
  fclose(f);

  *dim = (k / *N);

  f = fopen(fname, "r");
  x = malloc(*N * sizeof(double*));
  for (i=0; i<*N; i++) {
    x[i] = malloc(*dim * sizeof(double));
    for (j=0; j<*dim; j++) {
      k = fscanf(f, "%lg", &tmp);
      x[i][j] = tmp;
    }
  }
  fclose(f);

  return x;
}

int* load_discrete1(char*fname, int*dim, int*max) {
  FILE *f;
  int*x;
  int i,j,k;

  *dim = 0;
  *max = 0;

  k = 0;
  f = fopen(fname, "r");
  while (!feof(f)) {
    i = fscanf(f, "%d", &j);
    if (i > 0) { 
      k++; 
      if (j > *max) { *max = j; }
    }
  }
  *dim = k;
  fclose(f);

  x = malloc(*dim * sizeof(int));

  f = fopen(fname, "r");
  j = 0;
  while (!feof(f)) {
    i = fscanf(f, "%d", &k);
    if (i > 0) {
      x[j] = k-1;
      j++;
    }
  }
  fclose(f);

  return x;
}

int** load_discrete2(char*fname, int*dim1, int**dim2, int*max) {
  FILE *f;
  int**x;
  int*tmp;
  char*s;
  int i,j,k,l,maxlen;

  *max = 0;

  k = 0;
  f = fopen(fname, "r");
  while (!feof(f)) {
    i = fscanf(f, "%*[^\n]%*[\n]");
    k++;
  }
  fclose(f);
  *dim1 = k;

  x     = malloc(*dim1 * sizeof(int*));
  *dim2 = malloc(*dim1 * sizeof(int));

  s   = malloc(10000 * sizeof(char));
  tmp = malloc(10000 * sizeof(int));

  f = fopen(fname, "r");
  k = 0;
  while (!feof(f)) {
    l = 0;
    i = 1;
    while (1) {
      i = fscanf(f, "%[^ \n]", s);
      if (i <= 0) { break; }
      tmp[l] = atoi(s);
      if (tmp[l] > *max) { *max = tmp[l]; }
      l++;
      i = fscanf(f, "%[ \n]", s);
      if (s[0] == '\n') { break; }
    }
    x[k] = malloc(l * sizeof(int));
    for (j=0; j<l; j++) {
      x[k][j] = tmp[j] - 1;
    }
    (*dim2)[k] = l;
    k++;
  }
  fclose(f);

  free(s);
  free(tmp);

  return x;
}

double*   add_const_r_1 (double*   res, double cn, double*   vec, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = cn + vec[i];
  return res;
}

double**  add_const_r_2 (double**  res, double cn, double**  vec, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = cn + vec[i][j];
  return res;
}

double*** add_const_r_3 (double*** res, double cn, double*** vec, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = cn + vec[i][j][k];
  return res;
}

double*   mult_const_r_1 (double*   res, double cn, double*   vec, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = cn * vec[i];
  return res;
}

double**  mult_const_r_2 (double**  res, double cn, double**  vec, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = cn * vec[i][j];
  return res;
}

double*** mult_const_r_3 (double*** res, double cn, double*** vec, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = cn * vec[i][j][k];
  return res;
}

double*   add_vec_r_1   (double*   res, double*   vec1, double*   vec2, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = vec1[i] + vec2[i];
  return res;
}

double**  add_vec_r_2   (double**  res, double**  vec1, double**  vec2, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = vec1[i][j] + vec2[i][j];
  return res;
}

double*** add_vec_r_3   (double*** res, double*** vec1, double*** vec2, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = vec1[i][j][k] + vec2[i][j][k];
  return res;
}

double*   mult_vec_r_1   (double*   res, double*   vec1, double*   vec2, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = vec1[i] * vec2[i];
  return res;
}

double**  mult_vec_r_2   (double**  res, double**  vec1, double**  vec2, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = vec1[i][j] * vec2[i][j];
  return res;
}

double*** mult_vec_r_3   (double*** res, double*** vec1, double*** vec2, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = vec1[i][j][k] * vec2[i][j][k];
  return res;
}




int   *   add_const_i_1 (int   *   res, int    cn, int   *   vec, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = cn + vec[i];
  return res;
}

int   **  add_const_i_2 (int   **  res, int    cn, int   **  vec, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = cn + vec[i][j];
  return res;
}

int   *** add_const_i_3 (int   *** res, int    cn, int   *** vec, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = cn + vec[i][j][k];
  return res;
}

int   *   mult_const_i_1 (int   *   res, int    cn, int   *   vec, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = cn * vec[i];
  return res;
}

int   **  mult_const_i_2 (int   **  res, int    cn, int   **  vec, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = cn * vec[i][j];
  return res;
}

int   *** mult_const_i_3 (int   *** res, int    cn, int   *** vec, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = cn * vec[i][j][k];
  return res;
}

int   *   add_vec_i_1   (int   *   res, int   *   vec1, int   *   vec2, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = vec1[i] + vec2[i];
  return res;
}

int   **  add_vec_i_2   (int   **  res, int   **  vec1, int   **  vec2, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = vec1[i][j] + vec2[i][j];
  return res;
}

int   *** add_vec_i_3   (int   *** res, int   *** vec1, int   *** vec2, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = vec1[i][j][k] + vec2[i][j][k];
  return res;
}

int   *   mult_vec_i_1   (int   *   res, int   *   vec1, int   *   vec2, int lo1, int hi1) {
  int i;
  for (i=lo1-1;i<hi1;i++) res[i] = vec1[i] * vec2[i];
  return res;
}

int   **  mult_vec_i_2   (int   **  res, int   **  vec1, int   **  vec2, int lo1, int hi1, int lo2, int hi2) {
  int i,j;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) res[i][j] = vec1[i][j] * vec2[i][j];
  return res;
}

int   *** mult_vec_i_3   (int   *** res, int   *** vec1, int   *** vec2, int lo1, int hi1, int lo2, int hi2, int lo3, int hi3) {
  int i,j,k;
  for (i=lo1-1;i<hi1;i++) for (j=lo2-1;j<hi2;j++) for (k=lo3-1;k<hi3;k++) res[i][j][k] = vec1[i][j][k] * vec2[i][j][k];
  return res;
}


double sqrDiff(double*   a, double*   b, int lo1, int hi1) {
  double r;
  int i;
  r = 0;
  for (i=lo1-1;i<hi1;i++) r += (a[i] - b[i]) * (a[i] - b[i]);
  return r;
}
